
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gsl/gsl_sf_bessel.h>
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

#define LEN	2048
#define EPS	1e-9
#define SEG	16
#define BIS	gsl_sf_bessel_In_scaled

int	main(int argc,char **argv)
{
	int	Nx,Ny,N,y,x,n,ind,bps=(int)sizeof(Cell)*8;
	double	Ox,Oy,R,V,v,L,X,Y,*W,r,base,step,sum,div,pow,mul,inc,w,B=0.0;
	Cell	**C,**GF,c0,c1,c=(Cell)((1ULL<<bps)-1);
	char	id[LEN],*bvoc="bad value of circle.";

	if (argc!=10 && argc!=11)
	    Error("usage : c_gf Nx Ny Ox Oy R value {bias} L C.tif GF.tif");

	if ((Nx=atoi(argv[1]))<=0 ||
	    (Ny=atoi(argv[2]))<=0) Error("bad number of pixels.");

	Ox=atof(argv[3]);
	Oy=atof(argv[4]);

	if ((R=atof(argv[5]))<EPS) Error("bad radius of circle.");

	if (Ox-R<0.0 || Ox+R>(double)(Nx-1) ||
	    Oy-R<0.0 || Oy+R>(double)(Ny-1))
    (void)fputs("all or part of circle is out of image (warning).\n",stderr);

	if (argc==11)
	    if ((B=atof(argv[7]))<0.0 || B>(double)c) Error("bad bias value.");

	if ((V=atof(argv[6]))<0.0)
	    if (V+B<0.0) Error(bvoc); else v=B;
	else
	    if ((v=V+B)>(double)c) Error(bvoc);

	for (bps=1; (double)(1ULL<<bps)<=v; bps<<=1) ; c=(Cell)((1ULL<<bps)-1);

	if ((L=atof(argv[argc-3]))<EPS) Error("bad radius of Gaussian filter.");

	if ((X=(double)(Nx-1)-Ox)<Ox) X=Ox;
	if ((Y=(double)(Ny-1)-Oy)<Oy) Y=Oy;
	N=(int)(sqrt(X*X+Y*Y)*(double)SEG)+2;

	if ((W=(double *)malloc(sizeof(double)*N))==NULL ||
	    (C=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*C=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL ||
	    (GF=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*GF=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL)
	    Error("memory allocation error.");

	c0=(Cell)(B+0.5); c1=(Cell)(V+B+0.5);

	for (n=0; n<N; n++) {
	    if ((r=(double)n/(double)SEG)<R) {
		base=1.0; step=(-1.0); sum=BIS(0,mul); div=r/R;
	    }
	    else {
		base=0.0; step=1.0; sum=0.0; div=R/r;
	    }
	    pow=1.0; ind=0; mul=2.0*r*R/(L*L);

	    do sum+=(inc=(pow*=div)*BIS(++ind,mul)); while (fabs(inc/sum)>EPS);

	    W[n]=(base+step*exp(-(r-R)*(r-R)/(L*L))*sum)*V+B+0.5;
	}
	for (y=0; y<Ny; y++) { Y=(double)y-Oy;
	    if (y>0) {
		C[y]=C[y-1]+Nx; GF[y]=GF[y-1]+Nx;
	    }
	    for (x=0; x<Nx; x++) { X=(double)x-Ox;
		C[y][x]=((r=sqrt(X*X+Y*Y))>R)?c0:c1;

		n=(int)(w=r*(double)SEG);
		v=W[n]+(W[n+1]-W[n])*(w-(double)n);
		GF[y][x]=(v<0.0)?0:(v>(double)c)?c:(Cell)v;
	    }
	}
	(void)sprintf(id,"%lf\t%lf",-B,(double)c-B);

	if (argv[argc-2][0]!='-')
	    StoreImageFile(argv[argc-2],Nx,Ny,bps,C,id);

	if (argv[argc-1][0]!='-')
	    StoreImageFile(argv[argc-1],Nx,Ny,bps,GF,id);

	free(*GF); free(GF); free(*C); free(C); free(W);

	return 0;
}
