
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

#define LEN	2048
#define EPS	1e-9
#define SQRT_PI	1.77245385090551602728

int	main(int argc,char **argv)
{
	int	Nx,Ny,Nz,y,len,z,x,BPS=(int)sizeof(Cell)*8;
	double	Ox,Oy,Oz,R,V,v,L,RL,I0,Z,Y,X,r,m,p,I,B=0.0;
	Cell	**S,**GF,S0,S1,C=(Cell)((1ULL<<BPS)-1);
	char	F[LEN],ID[LEN],path[LEN],*bvos="bad value of sphere.";

	if (argc!=12 && argc!=13)
	    Error("usage : s_gf Nx Ny Nz Ox Oy Oz R value {bias} L S/ GF/");

	if ((Nx=atoi(argv[1]))<=0 ||
	    (Ny=atoi(argv[2]))<=0 ||
	    (Nz=atoi(argv[3]))<=0) Error("bad number of pixels.");

	Ox=atof(argv[4]);
	Oy=atof(argv[5]);
	Oz=atof(argv[6]);

	if ((R=atof(argv[7]))<0.0) Error("bad radius of sphere.");

	if (Ox-R<0.0 || Ox+R>(double)(Nx-1) ||
	    Oy-R<0.0 || Oy+R>(double)(Ny-1) ||
	    Oz-R<0.0 || Oz+R>(double)(Nz-1))
    (void)fputs("all or part of sphere is out of image (warning).\n",stderr);

	if (argc==13)
	    if ((B=atof(argv[9]))<0.0 || B>(double)C) Error("bad bias value.");

	if ((V=atof(argv[8]))<0.0)
	    if (V+B<0.0) Error(bvos); else v=B;
	else
	    if ((v=V+B)>(double)C) Error(bvos);

	for (BPS=1; (double)(1ULL<<BPS)<=v; BPS<<=1) ; C=(Cell)((1ULL<<BPS)-1);

	if ((L=atof(argv[argc-3]))<EPS) Error("bad radius of Gaussian filter.");

	if ((S=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*S=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL ||
	    (GF=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*GF=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL)
	    Error("memory allocation error.");

	for (y=1; y<Ny; y++) {
	    S[y]=S[y-1]+Nx; GF[y]=GF[y-1]+Nx;
	}
	S0=(Cell)(B+0.5); S1=(Cell)(V+B+0.5);

	RL=R/L; I0=erf(RL)-RL*exp(-RL*RL)*2.0/SQRT_PI;

	len=1; for (z=10; z<Nz; z*=10) ++len;
	(void)sprintf(F,"%%s/%%0%dd.tif",len);

	(void)sprintf(ID,"%lf\t%lf",-B,(double)C-B);

	for (z=0; z<Nz; z++) { Z=(double)z-Oz;
	    for (y=0; y<Ny; y++) { Y=(double)y-Oy;
	    for (x=0; x<Nx; x++) { X=(double)x-Ox;
		S[y][x]=((r=sqrt(X*X+Y*Y+Z*Z))>R)?S0:S1;

		if (r<EPS)
		    I=I0;
		else {
		    m=(R-r)/L;
		    p=(R+r)/L;
		    I=(erf(m)+erf(p)-L/r*(exp(-m*m)-exp(-p*p))/SQRT_PI)/2.0;
		}
		GF[y][x]=((v=I*V+B+0.5)<0.0)?0:(v>(double)C)?C:(Cell)v;
	    }}
	    if (argv[argc-2][0]!='-') {
		(void)sprintf(path,F,argv[argc-2],z);
		StoreImageFile(path,Nx,Ny,BPS,S,ID);
	    }
	    if (argv[argc-1][0]!='-') {
		(void)sprintf(path,F,argv[argc-1],z);
		StoreImageFile(path,Nx,Ny,BPS,GF,ID);
	    }
	}
	free(*GF); free(GF); free(*S); free(S);

	return 0;
}
